//
//  GMlib -- Graphics & Media Lab Common Source Library
//
//  $Id: gmlcamera.h,v 1.10 2004/04/02 08:43:58 leo Exp $
//
//  Copyright (C) 2004, Moscow State University Graphics & Media Lab
//  gmlsupport@graphics.cs.msu.su
//  
//  This file is part of GMlib software.
//  For conditions of distribution and use, see the accompanying README file.

/** @file gmlcamera.h
 *  @brief A class for 3d virtual camera
 * 
 */

#ifndef _GMLCAMERA_H_
#define _GMLCAMERA_H_

#include "../math/gmlmatrix4.h"
#include "gmlgenericcamera.h"

namespace gml
{
  /** @addtogroup Utils3D
   * @{
   */

  /// Class representing 3d virtual camera
class Camera: public GenericCamera
  {
    public:
      Camera()
      {
        m_view_angle = 90;
        m_aspect = 1.0;
        m_near = 1;
        m_far = 2;
        UpdateProjectionTransform();
      }

      /// Take attributes from the most generic camera
      Camera(const GenericCamera* gen_cam)
        {
        ASSERT(gen_cam);
        SetViewTransform(gen_cam->GetViewTransform());
        m_view_angle = 90;
        m_aspect = 1.0;
        m_near = 1;
        m_far = 2;
        UpdateProjectionTransform();
        };

      /// Set new clipping planes for the camera
      void SetClipPlanes(float p_near, float p_far)
        {
        m_near = p_near;
        m_far = p_far;
        UpdateProjectionTransform();
        }
     
      /// Return near clip plane
      float GetNearClipPlane() const
        { return m_near; }
  
      /// Return far clip plane
      float GetFarClipPlane() const
        { return m_far; }


      /// Set camera pixel aspect
      void SetAspect(float aspect)
        { 
        m_aspect = aspect; 
        UpdateProjectionTransform();
        }

      /// Return camera pixel aspect
      float GetAspect() const
        { return m_aspect;}

      /// set camera view angle (angle should be between 0 and 180 degrees)
      void SetViewAngle(float view_angle)
      {
        ASSERT(view_angle >= 0);
        ASSERT(view_angle <= 180);
        m_view_angle = view_angle;
       UpdateProjectionTransform();
      }
      /// return camera view angle
      float GetViewAngle() const
      {
        return m_view_angle;
      }

      /// set view transformation
      const Matrix4x4d& GetTransform() const
      {
        return GetViewTransform();
      }

      /// return view transformation
      void SetTransform(const Matrix4x4d& view_transform)
      {
        SetViewTransform(view_transform);
      }

      virtual GenericCamera* Clone() const
        {
        return new Camera(*this);
        }

    private:

      void UpdateProjectionTransform()
        {
         gml::Matrix4x4d persp_proj;
        persp_proj.Perspective(m_view_angle, m_aspect, m_near, m_far);
        SetProjectionTransform(persp_proj);
        }
    protected:

      float m_view_angle; ///< view angle (degrees)
      float m_aspect; ///< pixel aspect
      float m_near;  ///< near clip plane (
      float m_far; ///< far clip plane

  };

  /** @} */
} // namespace gml

#endif
