//
//  GMlib -- Graphics & Media Lab Common Source Library
//
//  $Id: gmlgenericcamera.h,v 1.3 2004/04/02 08:43:53 leo Exp $
//
//  Copyright (C) 2004, Moscow State University Graphics & Media Lab
//  gmlsupport@graphics.cs.msu.su
//  
//  This file is part of GMlib software.
//  For conditions of distribution and use, see the accompanying README file.

/** @file gmlgenericcamera.h
 *  @brief A class for generic 3d virtual camera
 * 
 */

#ifndef _GMLGENERICCAMERA_H_
#define _GMLGENERICCAMERA_H_

#include "../math/gmlmatrix4.h"
#include "../math/gmlray3.h"
#include "../base/gmlsmartobject.h"

namespace gml
  {
  /** @addtogroup Utils3D
   * @{
   */

  /// Class representing any 3d virtual camera (view and projection matrices)
  class GenericCamera : public SmartObject
    {
    public:

      GenericCamera()
        {
        m_view_transform.MakeIdentity();
        m_inv_view_transform.MakeIdentity();
        m_proj_transform.MakeIdentity();
        m_inv_proj_transform.MakeIdentity();
        }

      /// get view vector of camera (normalized)
      Vector3d GetViewVector() const
        {
        Vector3d dir(0, 0, -1);
        m_inv_view_transform.MultMatrixDir(dir);
        return dir.Normalize();
        }

      /// get up vector of camera (normalized)
      Vector3d GetUpVector() const
        {
        Vector3d dir(0, 1, 0);
        m_inv_view_transform.MultMatrixDir(dir);
        return dir.Normalize();
        }

      /// get up vector of camera (normalized)
      Vector3d GetRightVector() const
        {
        Vector3d dir(1, 0, 0);
        m_inv_view_transform.MultMatrixDir(dir);
        return dir.Normalize();
        }

      // get camera position
      Vector3d GetObserver() const
        {
        return m_inv_view_transform.GetTranslation();
        }

      void SetObserver(const Vector3d& obs)
        {
        m_inv_view_transform.SetTranslation(obs);
        m_view_transform = m_inv_view_transform.Inverse();
        }

      Matrix4x4d GetRotation() const
        {
        return m_inv_view_transform.GetRotation();
        }

      void SetRotation(const Matrix4x4d& rot)
        {
        m_inv_view_transform.SetRotation(rot);
        m_view_transform = m_inv_view_transform.Inverse();
        }

      const Matrix4x4d& GetViewTransform() const
        {
        return m_view_transform;
        }

      const Matrix4x4d& GetProjectionTransform() const
        {
        return m_proj_transform;
        }

      /// Virtual duplication (allow to create copies of derived classes)
      //  @attention Reimplement if you want to clone derived classes!
      virtual GenericCamera* Clone() const
        {
        return new GenericCamera(*this);
        }

      /// Restore eye coordinates for given point in clip (normalized) coordinates
      void UnProjectToEye(const gml::Vector3d& point, gml::Vector3d& rest_point) const
        {
        m_inv_proj_transform.MultMatrixVec(point, rest_point);
        }

      /// Restore eye coordinates for given point in clip (normalized) coordinates
      void UnProjectToEye(gml::Vector3d& point) const
        {
        UnProjectToEye(gml::Vector3d(point), point);
        }

      /// Restore world coordinates from given point in clip coordinates
      void UnProject(gml::Vector3d& point) const
        {
        UnProjectToEye(point);
        m_inv_view_transform.MultMatrixVec(point);
        };

      /// Restore world coordinates from given point in clip coordinates
      void UnProject(gml::Vector3d& point, gml::Vector3d& rest_point)
        {
        UnProjectToEye(point, rest_point);
        m_inv_view_transform.MultMatrixVec(rest_point);
        }

      /// Restore frustum corners at given distance
      /// @note points order: (near rect, far rect), 
      /// each rect's order: lower left, lower right, top right, top left
      void GetFrustum(gml::Vector3d points[8])
        {

        points[0].SetValue(-1, -1, -1);
        UnProject(points[0]);

        points[1].SetValue(1, -1, -1);
        UnProject(points[1]);

        points[2].SetValue(1, 1, -1);
        UnProject(points[2]);

        points[3].SetValue(-1, 1, -1);
        UnProject(points[3]);

        points[4].SetValue(-1, -1, 1);
        UnProject(points[4]);

        points[5].SetValue(1, -1, 1);
        UnProject(points[5]);

        points[6].SetValue(1, 1, 1);
        UnProject(points[6]);

        points[7].SetValue(-1, 1, 1);
        UnProject(points[7]);
        }


      /// Throw a ray trough the (x,y) location of the canonical 3d viewport
      void CastRay(double x, double y, gml::Ray3& out_ray) const
        {
        gml::Vector3d p1(x, y, -1);
        gml::Vector3d p2(x, y, 1);
        UnProject(p1);
        UnProject(p2);
        out_ray.SetValue(p1, (p2 - p1).Normalize());
        }

      /// Return a point (out_x,out_y), where given ray intersects a canonical 3d viewport
      void IntersectNearPlaneWithRay(const gml::Ray3& in_ray, double& out_x, double& out_y)
        {
        // transform the ray into canonical viewport
        gml::Ray3 tr_ray;
        in_ray.GetTransformedRay(m_view_transform, tr_ray);
        double t = - in_ray.GetStart().z / in_ray.GetDir().z;
        gml::Vector3d vct = in_ray.GetRayPos(t);
        out_x = vct.x;
        out_y = vct.y;
        }

    protected:

      void SetViewTransform(const Matrix4x4d& view_transform)
        {
        m_view_transform = view_transform;
        m_inv_view_transform = m_view_transform.Inverse();
        }

      void SetProjectionTransform(const Matrix4x4d& proj_transform)
        {
        m_proj_transform = proj_transform;
        m_inv_proj_transform = m_proj_transform.Inverse();
        }

    protected:

      Matrix4x4d m_view_transform; ///< OGL ready viewing transform (global -> eye)
      Matrix4x4d m_inv_view_transform; ///< inverse view transform (eye -> global)
      Matrix4x4d m_proj_transform; ///< OGL ready projection transform (eye -> clip)
      Matrix4x4d m_inv_proj_transform; ///< inverse projection transform (clip -> eye)
    };

  /** @} */
  } // namespace gml

#endif
