#ifndef _ANIMDATA_H_
#define _ANIMDATA_H_

// std includes
#include <vector>

// csl includes
#include "../../csl/base/gmlsmartobject.h"
#include "../../csl/base/gmlref.h"
#include "../../csl/math/gmlvector3.h"
#include "../../csl/utils3d/gmlcamera.h"

namespace pug
  {
  class xml_node;
  }

namespace cganim
  {

  // forwards
  class  AnimObject;

  /**

  @brief Contains animation data as well as loading facility

  Usage:

  @code 

  // include GML lib required include
  #include "../csl/base/gmlcommon.h"
  
    // include animdata headers
    #include "../animdata/include/animdata.h"
    #include "../animdata/include/exception.h"
    #include "../animdata/include/sphere.h"
    #include "../animdata/include/rectangle.h"
    
    using namespace gml;
    using namespace cganim;
 
   ...

   AnimationData* anim_data = new AnimationData;
  
  try
    {
    anim_data->Load("..\\demos\\simple.cganim");
    }
  catch(IOException& ex)
    {
    printf("Loading error: %s", ex.What().c_str());
    return -1;
    }

  for(int i = 0; i < anim_data->GetNObjects(); ++i)
    {
    AnimObject* obj = anim_data->GetAnimObject(i);

    Sphere* sphere = dynamic_cast<Sphere*>(obj);
    if (sphere != NULL)
      {
      printf("Sphere object found\n");
      continue;
      }

    Rectangle* rect = dynamic_cast<Rectangle*>(obj);
    if (rect != NULL)
      {
      printf("Rect object found\n");
      continue;
      };
    }
  delete anim_data;

  @endcode

  **/
  class AnimationData: public gml::SmartObject
    {
    public:
      
      /// Default constructor
      AnimationData();

      /// Load source animation data from given filename
      void Load(const std::string& filename);
      
      /// Return true if the file is already loaded
      bool IsLoaded() const;


      /// Return total number of objects
      int GetNObjects() const;
      
      /// Return i-th object
      const AnimObject* GetAnimObject(int i) const;

      /// Return i-th object
      AnimObject* GetAnimObject(int i);

      /// Return camera
      gml::Camera* GetCamera();

      /// Return number of lights in the scene
      int GetNLights() const;

      /// Return position of i-th point light
      gml::Vector3d GetLightPosition(int i) const;

    private:

      void ParseXML(pug::xml_node node);
      void ParseSphere(pug::xml_node node);
      void ParseRectangle(pug::xml_node node);
      void ParseCamera(pug::xml_node node);
      void ParseLight(pug::xml_node node);

      //void ParseParticle(pug::xml_node node);
      //void ParseEnvironment(pug::xml_node node);

      /// Find and update attibutes common to all anim objects
      void ParseAnimObjectCommonAttributes(pug::xml_node node, AnimObject* obj);

      /// Form a vector from "X Y Z" string
      gml::Vector3d ParseVector(const std::string& str);
      /// Find a "velocity" attribute and convert it to the vector
      bool ParseVelocity(pug::xml_node node, gml::Vector3d& out_velocity);
      /// Find a "mass" attribute and convert it to double
      bool ParseMass(pug::xml_node node, double& out_mass);
      /// Find a "color" attribute and convert to a vector
      bool ParseColor(pug::xml_node node, gml::Vector3d& out_color); 
      /// Find a 'texture" attribute and convert it to a string
      bool ParseTexture(pug::xml_node node, std::string& out_tex_path);
      /// Find a 'matte' attribute and convert it to a double
      bool ParseSemiTransp(pug::xml_node node, double& out_matte);
      /// Find a 'specular' attribute and convert it to a double
      bool ParseSpecular(pug::xml_node node, double& out_specular);
    private:

      /// True if some file was already opened successfully
      bool m_loaded;

      /// Array filled from parsed XML-file
      std::vector<gml::Ref<AnimObject> > m_loaded_objects;

      /// Camera 
      gml::Ref<gml::Camera> m_camera;

      /// Keep point lights positions here
      std::vector<gml::Vector3d> m_lights;
    };
  }

#endif