#ifndef _ANIM_OBJECT_H_
#define _ANIM_OBJECT_H_

// std includes
#include <string>

// csl includes

#include "../../csl/base/gmlsmartobject.h"
#include "../../csl/math/gmlvector3.h"


namespace cganim
  {
  
  /**

   @brief Base class for all geometry objects in animated scene. Contains common attributes.

  **/
  class AnimObject: public gml::SmartObject
    {
    public:

    /// Default constuctor. Default mass is 1kg, speed 0m/s
    AnimObject()
      : m_velocity(0.0), m_mass(1), m_color(0.7, 0.7, 0.7), m_ks(0), m_matte(1)
      {}

    /// Get object velocity (m/sec)
    gml::Vector3d GetVelocity() const
      { 
      return m_velocity; 
      }

    /// Set new object velocity (m/sec)
    void SetVelocity(const gml::Vector3d& new_velocity)
      {
      m_velocity = new_velocity;
      }

    /// Return object mass (kg)
    double GetMass() const
      {
      return m_mass;
      }

    /// Set new object mass (kg)
    void SetMass(double new_mass)
      {
      m_mass = new_mass;
      }

    /// Return object color
    gml::Vector3d GetColor() const
      {
      return m_color;
      }

    /// Modify object color
    void SetColor(const gml::Vector3d& new_color)
      {
      m_color = new_color;
      }

    /// Return object texture file path. Can be empty
    std::string GetTexture() const
      {
      return m_texture;
      }

    /// Modify texture path of the object
    void SetTexture(const std::string& texture)
      {
      m_texture = texture;
      }

    /// Return level of specularity for this object, 0 - non-specular, 1 - ideal mirror
    double GetSpecularLevel() const
      {
      return m_ks;
      }

    /// Modify level of specularity for this object, 0 - non-specular, 1 - ideal mirror
    void SetSpecularLevel(double ks)
      {
      m_ks = ks;
      }

    /// Set new transparency coefficient (0 - fully transparent, 1 - opaque)
    void SetMatte(double matte)
      {
      m_matte = matte;
      }

    /// Get transp coefficient (0 - fully transparent, 1 - opaque)
    double GetMatte() const
      {
      return m_matte;
      }

    private:

      /// speed of object (m/sec)
      gml::Vector3d m_velocity;

      /// mass of object (kg)
      double m_mass;

      /// color of object (RGB, [0..1])
      gml::Vector3d m_color;


      /// Path to texture
      std::string m_texture;

      /// Specular level (0 - diffuse, 1 - ideal mirror)
      double m_ks;

      /// Opaque level ( 0 - fully transparent, 1 - opaque)
      double m_matte;
    };

  }
#endif