#include "stdafx.h"

// gml include
#include <csl/files/gmlpathstring.h>
#include <csl/math/gmlmatrix4.h>
#include <csl/math/gmlquat.h>
// class include
#include "../include/animdata.h"

// own includes
#include "../include/exception.h"
#include "../include/animobject.h"
#include "../include/sphere.h"
#include "../include/rectangle.h"

using namespace cganim;

//////////////////////////////////////////////////////////////////////////////
/// 
//////////////////////////////////////////////////////////////////////////////
AnimationData::AnimationData()
: m_loaded(false)
  {
  }
//////////////////////////////////////////////////////////////////////////////
/// @exception cganim::IOException if file cannot be loaded
//////////////////////////////////////////////////////////////////////////////
void AnimationData::Load(const std::string& filename)
  {
  
  if (m_loaded)
    throw LogicException("Already opened");

  gml::PathString file_path(filename);

  if (!file_path.IsFile())
    throw IOException("Not a file");

  pug::xml_parser* xml_parser = new pug::xml_parser;

  bool res = xml_parser->parse_file(filename.c_str());
  if (!res)
    throw IOException("Can't open file");

  // set loaded flag
  m_loaded = true;

  // process loaded data
  ParseXML(xml_parser->document());

  delete xml_parser;
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
bool AnimationData::IsLoaded() const
  {
  return m_loaded;
  }


//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
int AnimationData::GetNObjects() const
  {
  if (!m_loaded)
    throw LogicException("Not loaded");
  return m_loaded_objects.size();
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
const AnimObject* AnimationData::GetAnimObject(int i) const
  {
  if (!m_loaded)
    throw LogicException("Not loaded");
  if (i < 0 || i >= m_loaded_objects.size())
    throw LogicException("Invalid index");

  return m_loaded_objects[i];
  }


//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
AnimObject* AnimationData::GetAnimObject(int i)
  {
  if (!m_loaded)
    throw LogicException("Not loaded");
  if (i < 0 || i >= m_loaded_objects.size())
    throw LogicException("Invalid index");

  return m_loaded_objects[i];
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
gml::Camera* AnimationData::GetCamera()
  {
  if (!m_loaded)
    throw LogicException("Not loaded");

  return m_camera;
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
int AnimationData::GetNLights() const
  {
  return m_lights.size();
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
gml::Vector3d AnimationData::GetLightPosition(int i) const
  {
  return m_lights[i];
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
void AnimationData::ParseXML(pug::xml_node node)
  {

  int n_children = node.children();
  if (n_children != 1)
    throw InvalidFileException("No root tag");
  
  // get access to the root node
  pug::xml_node anim_objects_node = node.child(0);

  std::string name = anim_objects_node.name();
  if (name != "anim_objects")
     throw InvalidFileException("Invalid root tag");

  // iterate by children
  pug::xml_node::xml_node_iterator iter = anim_objects_node.children_begin();
  for(; iter != anim_objects_node.children_end(); ++iter)
    {
    pug::xml_node child_node = *iter;

    std::string node_name = child_node.name();
    if (node_name == "sphere")
      {
      ParseSphere(child_node);
      };

    if (node_name == "rect")
      {
      ParseRectangle(child_node);
      };

    if (node_name == "camera")
      {
      ParseCamera(child_node);
      }
    
    if (node_name == "light")
      {
      ParseLight(child_node);
      }
    }
  }

//////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
 void AnimationData::ParseSphere(pug::xml_node node)
   {

   gml::Ref<Sphere> sphere = new Sphere;
   
   // parse radius
   if (node.has_attribute("radius"))
     {
     pug::xml_attribute attr = node.attribute("radius");
     
     double radius = attr;

     if (radius < 0)
      throw InvalidFileException("Invalid sphere radius attribute");

     sphere->SetRadius(radius);
     }

   // parse center
   if (node.has_attribute("center"))
     {
     pug::xml_attribute attr = node.attribute("center");
     
     gml::Vector3d position;
     std::string attr_string = attr;
     
     position = ParseVector(attr_string);
     
     sphere->SetCenter(position);
     }
   
   ParseAnimObjectCommonAttributes(node, sphere);

   m_loaded_objects.push_back(gml::ConvRef<AnimObject>(sphere));
   }

 //////////////////////////////////////////////////////////////////////////////
///
//////////////////////////////////////////////////////////////////////////////
 void AnimationData::ParseRectangle(pug::xml_node node)
   {

   gml::Ref<Rectangle> rect = new Rectangle;

   for(int i = 1; i <= 4; ++i)
     {
     // form corner attr name
     gml::String attr_name;
     attr_name.Format("p%d",i);
     
     // parse corner
     if (node.has_attribute(attr_name))
       {
       pug::xml_attribute attr = node.attribute(attr_name);
       
       gml::Vector3d position;
       std::string attr_string = attr;
     
       position = ParseVector(attr_string);


       // update corner position
       rect->SetCorner(i - 1, position);

       }
     }
 
   ParseAnimObjectCommonAttributes(node, rect);

   m_loaded_objects.push_back(gml::ConvRef<AnimObject>(rect));
   }

 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////

 void AnimationData::ParseCamera(pug::xml_node node)
   {
   gml::Ref<gml::Camera> camera = new gml::Camera; 
   
   gml::Vector3d position;
   gml::Vector3d target;
   gml::Vector3d up;
   double angle;
   double near_plane, far_plane;
   
   // position
   
   if (node.has_attribute("position"))
     {
     pug::xml_attribute attr = node.attribute("position");
     
     std::string attr_string = attr;
     
     position = ParseVector(attr_string);
     }
   
   if (node.has_attribute("target"))
     {
     pug::xml_attribute attr = node.attribute("target");
     
     std::string attr_string = attr;
     
     target = ParseVector(attr_string);
     
     }
   
   if (node.has_attribute("up"))
     {
     pug::xml_attribute attr = node.attribute("up");
     
     std::string attr_string = attr;
     
     up = ParseVector(attr_string);
     
     }
   
   if (node.has_attribute("angle"))
     {
     pug::xml_attribute attr = node.attribute("angle");
     
     angle = attr;
     
     }
   
   if (node.has_attribute("near"))
     {
     pug::xml_attribute attr = node.attribute("near");
     
     near_plane = attr;
     
     }
   
   if (node.has_attribute("far"))
     {
     pug::xml_attribute attr = node.attribute("far");
     
     far_plane = attr;
     
     }

   /*
   gml::Matrix4x4d rotation;
   gml::Quaterniond quat;
   quat.SetValue(gml::Vector3d(0, 0, -1), gml::Vector3d(0, 1, 0), (target - position).Normalize(), up);
   //quat.SetValue(gml::Vector3d(0, 0, -1), gml::Vector3d(0, 1, 0), gml::Vector3d(0, 0, 1), up);
   quat.GetValue(rotation);
   */

   gml::Vector3d view = (target - position).Normalize();
   gml::Vector3d up_corr = view.OrthogonalInGivenHalfplane(up);
   gml::Vector3d right = gml::CrossProduct(view, up);
   gml::Matrix4x4d rotation(1);
   rotation.SetColumn(0, gml::Vector4d(right, 0));
   rotation.SetColumn(1, gml::Vector4d(up_corr, 0));
   rotation.SetColumn(2, gml::Vector4d(-view, 0));

   camera->SetObserver(position);
   camera->SetViewAngle(angle);
   camera->SetClipPlanes(near_plane, far_plane);
   camera->SetRotation(rotation);

   m_camera = camera;
   }

  //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////

 void AnimationData::ParseLight(pug::xml_node node)
   {
   if (node.has_attribute("position"))
     {
     pug::xml_attribute attr = node.attribute("position");
     
     std::string attr_string = attr;
     
     gml::Vector3d position = ParseVector(attr_string);

     m_lights.push_back(position);
     }
   }


 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 gml::Vector3d AnimationData::ParseVector(const std::string& str)
   {
   gml::Vector3d position(0, 0, 0);

   int n_scanned = sscanf(str.c_str(), "%lf %lf %lf", &position.x, &position.y, &position.z);
   if (n_scanned != 3)
     throw InvalidFileException("Invalid vector attribute");

   return position;
   }
 
 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 
 bool AnimationData::ParseVelocity(pug::xml_node node, gml::Vector3d& out_velocity)
   {
   if (!node.has_attribute("velocity"))
     return false;

   pug::xml_attribute attr = node.attribute("velocity");
   
   std::string attr_string = attr;
   
   out_velocity = ParseVector(attr_string);

   return true;
   }
 
 
 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 
 bool AnimationData::ParseColor(pug::xml_node node, gml::Vector3d& out_color)
   {
   if (!node.has_attribute("color"))
     return false;

   pug::xml_attribute attr = node.attribute("color");
   
   std::string attr_string = attr;
   
   gml::Vector3d color = ParseVector(attr_string);
   if (color.MaxValue() > 1 || (-color).MaxValue() > 0)
     throw InvalidFileException("Invalid color");

   out_color  = color;

   return true;
   }

 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 
 bool AnimationData::ParseMass(pug::xml_node node, double& out_mass)
   {
   if (!node.has_attribute("mass"))
     return false;

   pug::xml_attribute attr = node.attribute("mass");
   
   double mass = attr;
   if (mass < 0)
     throw InvalidFileException("Invalid mass attribute");

   out_mass = attr;

   return true;
   }
 
 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 bool AnimationData::ParseTexture(pug::xml_node node, std::string& out_tex_path)
   {
    if (!node.has_attribute("texture"))
     return false;

   out_tex_path = (const char*)node.attribute("texture");
  
   return true;
   }

 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 
 bool AnimationData::ParseSemiTransp(pug::xml_node node, double& out_matte)
   {
   if (!node.has_attribute("matte"))
     return false;
   
   pug::xml_attribute attr = node.attribute("matte");
   
   double matte = attr;
   if (matte < 0 || matte > 1)
     throw InvalidFileException("Invalid matte attribute");
   
   out_matte = attr;
   
   return true;
   }
 
 
 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 
 bool AnimationData::ParseSpecular(pug::xml_node node, double& out_specular)
   {
   if (!node.has_attribute("specular"))
     return false;
   
   pug::xml_attribute attr = node.attribute("specular");
   
   double specular = attr;
   if (specular < 0 || specular > 1)
     throw InvalidFileException("Invalid specular attribute");
   
   out_specular = attr;
   
   return true;
   }
 //////////////////////////////////////////////////////////////////////////////
 ///
 //////////////////////////////////////////////////////////////////////////////
 void AnimationData::ParseAnimObjectCommonAttributes(pug::xml_node node, AnimObject* obj)
   {
    // parse velocity
   gml::Vector3d velocity;
   bool is_vel = ParseVelocity(node, velocity);
   if (is_vel)
     obj->SetVelocity(velocity);

   // parse mass
   double mass;
   bool is_mass = ParseMass(node, mass);
   if (is_mass)
     obj->SetMass(mass);


    // parse color
   gml::Vector3d color;
   bool is_color = ParseColor(node, color);
   if (is_color)
     obj->SetColor(color);

    // parse texture
   std::string tex_path;
   bool is_texture = ParseTexture(node, tex_path);
   if (is_texture)
     obj->SetTexture(tex_path);

    // parse matte
   double matte;
   bool is_matte = ParseSemiTransp(node, matte);
   if (is_matte)
     obj->SetMatte(matte);

   // parse specular
   double specular;
   bool is_specular = ParseSpecular(node, specular);
   if (is_specular)
     obj->SetSpecularLevel(specular);
   }


